/********************************************************************** 
 Freeciv - Copyright (C) 2003 - M. Kaufman
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

/**********************************************************************
 This is an extremely simple database lookup for Freeciv authentication.
 The username and password along with the proposed action are passed in 
 on argv and the exit status gives the caller the action taken.

 The arguments are:

 0: program name
 1: username
 2: password
 3: action 
     0 = add new username and password to database.
     1 = check username and password against the one in database.

 The return statuses are:

 1: the wrong number of required arguments,
 2: an error occurred, the action byte was not '0' or '1', or we couldn't
    access the database file.
 3: the password we checked for was ok, or we were able to successfully
    insert an entry.
 4: the entry we were checking for was not found.
 5: the entry's password was wrong.

 Industrial strength solutions probably demand a real database.
 If you want information about such solutions, contact 
 freeciv-dev@freeciv.org
***********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "user_db.h" /* for enum */

#define FC_USER_PROG "freeciv_user_database"
#define FC_LOCK      "freeciv_user_database.lock"

#ifdef USER_DB_PATH
  #define FC_USER_DATABASE USER_DB_PATH "/" FC_USER_PROG
  #define FC_LOCK_FILE USER_DB_PATH "/" FC_LOCK
#else
  #define FC_USER_DATABASE FC_USER_PROG
  #define FC_LOCK_FILE FC_LOCK
#endif

#define MAX_LEN_FIELD 	32   /* for parity with Freeciv's MAX_LEN_NAME */
#define SLEEP_TIME	50   /* us to sleep after finding a lock */
#define SLEEP_MAX	100  /* max times to sleep before declaring failure */

static enum userdb_status insert_new_entry(char *username, char *password);
static enum userdb_status check_entry(char *username, char *password);

static int check_lock(void);
static void remove_lock(void);

static int we_locked;

/**************************************************************************
  parse command line parameters, call lookup and return exit status
**************************************************************************/
int main(int argc, char **argv)
{
  char username[MAX_LEN_FIELD], password[MAX_LEN_FIELD];
  enum userdb_status s = USER_DB_ERROR;

  /* check if all the required arguments are here */
  if (argc != 4) {
    exit(USER_DB_BAD_INPUT);
  }

  strncpy(username, argv[1], MAX_LEN_FIELD);
  strncpy(password, argv[2], MAX_LEN_FIELD);

  /* insert new entry */
  if (argv[3][0] == '0') {
    s = insert_new_entry(username, password);
  } else if (argv[3][0] == '1') { /* search for existant entry to compare */
    s = check_entry(username, password);
  }

  remove_lock();
  exit(s);

  return 0;
}

/**************************************************************************
 create a lock, so as other servers aren't accessing the database when we are.
 return 1 for success, 0 for failure
**************************************************************************/
static int check_lock(void)
{
  FILE *lock;
  static int check;

  lock = fopen(FC_LOCK_FILE, "r");

  /* if we aren't locked, create one */
  if (!lock) {
    lock = fopen(FC_LOCK_FILE, "w");
    fclose(lock);

    we_locked = 1;
    check = 0;
    return 1;
  } else {
    /* we found a lock, sleep a while and check again */
    fclose(lock);
    usleep(SLEEP_TIME);

    /* check if we've done this, don't wait forever, give up if need be */
    if (check++ > SLEEP_MAX) {
      check = 0;
      return 0;
    }

    check_lock();
  }
}

/**************************************************************************
 remove the lock file.
**************************************************************************/
static void remove_lock(void)
{
  if (we_locked) {
    remove(FC_LOCK_FILE);
    we_locked = 0;
  }
}

/**************************************************************************
 append a username and password to the database. Does not check if the
 entry already exists.
**************************************************************************/
static enum userdb_status insert_new_entry(char *username, char *password)
{
  FILE *fp;

  if (!check_lock()) {
    return USER_DB_ERROR;
  }

  fp = fopen(FC_USER_DATABASE, "a"); 

  if (!fp) {
    return USER_DB_ERROR;
  }

  /* insert */
  fprintf(fp, "%s\t%s\n", username, password);

  fclose(fp);

  return USER_DB_SUCCESS;  
}

/**************************************************************************
 check the parameters against an entry that should exit in the database
**************************************************************************/
static enum userdb_status check_entry(char *username, char *password)
{
  FILE *fp;
  char file_user[MAX_LEN_FIELD], file_pass[MAX_LEN_FIELD];

  if (!check_lock()) {
    return USER_DB_ERROR;
  }

  fp = fopen(FC_USER_DATABASE, "r"); 

  if (!fp) {
    fp = fopen(FC_USER_DATABASE, "w");

    if (!fp) {
      return USER_DB_ERROR; /* can't open it period */
    } else {
      fclose(fp);
      return USER_DB_NOT_FOUND; /* we needed to create the database */
    }
  }

  /* scan through each line until end of file looking for the username */
  while (fscanf(fp, "%s\t%s", file_user, file_pass) != EOF) {
    if (strncmp(username, file_user, MAX_LEN_FIELD) == 0) {
      if (password[0] == '\0') {
        return USER_DB_FOUND;
      } else if (strncmp(password, file_pass, MAX_LEN_FIELD) == 0) {
        return USER_DB_SUCCESS;
      } else {
        return USER_DB_PASS_WRONG;
      }
    }
  }

  fclose(fp);

  return USER_DB_NOT_FOUND;  
}
